# Modified from:
#   DiT:  https://github.com/facebookresearch/DiT/blob/main/sample.py
import torch
torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.allow_tf32 = True
torch.set_float32_matmul_precision('high')
setattr(torch.nn.Linear, 'reset_parameters', lambda self: None)
setattr(torch.nn.LayerNorm, 'reset_parameters', lambda self: None)
from torchvision.utils import save_image
import os
import numpy as np
from tqdm import tqdm
import random

import sys
sys.path.append(".")

import time
import argparse
from tokenizer.tokenizer_image.vq_model import VQ_models
from autoregressive.models.modeling_llama import LlamaForCausalLM
from torch.nn import functional as F

def cfg_logit_process(combined_logits, cfg_scale=4.0, cfg_interval=-1):
    cond_logits, uncond_logits = torch.split(combined_logits, len(combined_logits) // 2, dim=0)
    logits = uncond_logits + (cond_logits - uncond_logits) * cfg_scale
    return logits


def top_k_top_p_filtering(
    logits,
    top_k: int = 0,
    top_p: float = 1.0,
    filter_value: float = -float("Inf"),
    min_tokens_to_keep: int = 1,
):
    """Filter a distribution of logits using top-k and/or nucleus (top-p) filtering
    Args:
        logits: logits distribution shape (batch size, vocabulary size)
        if top_k > 0: keep only top k tokens with highest probability (top-k filtering).
        if top_p < 1.0: keep the top tokens with cumulative probability >= top_p (nucleus filtering).
            Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751)
        Make sure we keep at least min_tokens_to_keep per batch example in the output
    From: https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317
    """
    if top_k > 0:
        top_k = min(max(top_k, min_tokens_to_keep), logits.size(-1))  # Safety check
        # Remove all tokens with a probability less than the last token of the top-k
        indices_to_remove = logits < torch.topk(logits, top_k)[0][..., -1, None]
        logits[indices_to_remove] = filter_value

    if top_p < 1.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

        # Remove tokens with cumulative probability above the threshold (token with 0 are kept)
        sorted_indices_to_remove = cumulative_probs > top_p
        if min_tokens_to_keep > 1:
            # Keep at least min_tokens_to_keep (set to min_tokens_to_keep-1 because we add the first one below)
            sorted_indices_to_remove[..., :min_tokens_to_keep] = 0
        # Shift the indices to the right to keep also the first token above the threshold
        sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
        sorted_indices_to_remove[..., 0] = 0

        # scatter sorted tensors to original indexing
        indices_to_remove = sorted_indices_to_remove.scatter(1, sorted_indices, sorted_indices_to_remove)
        logits[indices_to_remove] = filter_value
    return logits


def sample(logits, temperature: float=1.0, top_k: int=0, top_p: float=1.0, sample_logits=True):        
    logits = logits[:, -1, :] / max(temperature, 1e-5)
    if top_k > 0 or top_p < 1.0:
        logits = top_k_top_p_filtering(logits, top_k=top_k, top_p=top_p)
    probs = F.softmax(logits, dim=-1)
    if sample_logits:
        idx = torch.multinomial(probs, num_samples=1)
    else:
        _, idx = torch.topk(probs, k=1, dim=-1)
    return idx, probs

def generate(model, c_indices, max_new_tokens, cfg_scale=4.0, cfg_interval=-1, **sampling_kwargs):
    model.eval()
    with torch.no_grad():
        device = c_indices.device
        max_batch_size = c_indices.shape[0]
        
        # Prepare the conditional combined input once
        if model.model.model_type == 'c2i':
            if cfg_scale > 1.0:
                cond_null = torch.ones_like(c_indices) * model.model.num_classes
                cond_combined = torch.cat([c_indices, cond_null])
            else:
                cond_combined = c_indices
        else:
            raise NotImplementedError(f"input type {model.model.input_type} is not supported")
        
        # Preallocate the sequence tensor
        seq = torch.empty((max_batch_size, max_new_tokens), dtype=torch.int, device=device)
        
        # Initial output
        output = model(cond_idx=cond_combined)
        past_key_values = output.past_key_values
        
        if cfg_scale > 1.0:
            cond_logits, uncond_logits = torch.split(output.logits, max_batch_size, dim=0)
            logits = uncond_logits + (cond_logits - uncond_logits) * cfg_scale
        else:
            logits = output.logits
        
        next_token, _ = sample(logits, **sampling_kwargs)
        seq[:, 0] = next_token.squeeze(-1)
        
        # Iteratively generate tokens
        for i in range(1, max_new_tokens):
            input_ids = next_token.expand(2, -1)
            output = model(input_ids=input_ids, past_key_values=past_key_values)
            past_key_values = output.past_key_values
            
            if cfg_scale > 1.0:
                cond_logits, uncond_logits = torch.split(output.logits, max_batch_size, dim=0)
                logits = uncond_logits + (cond_logits - uncond_logits) * cfg_scale
            else:
                logits = output.logits
            
            next_token, _ = sample(logits, **sampling_kwargs)
            seq[:, i] = next_token.squeeze(-1)
            
    return seq


def main(args):
    # Setup PyTorch:
    torch.manual_seed(args.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.set_grad_enabled(False)
    device = "cuda" if torch.cuda.is_available() else "cpu"

    # create and load model
    vq_model = VQ_models[args.vq_model](
        codebook_size=args.codebook_size,
        codebook_embed_dim=args.codebook_embed_dim)
    vq_model.to(device)
    vq_model.eval()
    checkpoint = torch.load(args.vq_ckpt, map_location="cpu")
    vq_model.load_state_dict(checkpoint["model"])
    del checkpoint
    print(f"image tokenizer is loaded")

    # create and load gpt model
    precision = {'none': torch.float32, 'bf16': torch.bfloat16, 'fp16': torch.float16}[args.precision]
    latent_size = args.image_size // args.downsample_size
    gpt_model = LlamaForCausalLM.from_pretrained(
        args.gpt_ckpt,
    ).to(device=device, dtype=precision)
    
    print(f"gpt model is loaded")

    if args.compile:
        print(f"compiling the model...")
        gpt_model = torch.compile(
            gpt_model,
            mode="reduce-overhead",
            fullgraph=True
        ) # requires PyTorch 2.0 (optional)
    else:
        print(f"no need to compile model in demo") 

    # Labels to condition the model with (feel free to change):
    # class_labels = [207, 360, 387, 974, 88, 979, 417, 279]
    
    os.makedirs(f"sample_idx/{os.path.basename(args.gpt_ckpt)}", exist_ok=True)
    for idx in range(args.start_idx, args.end_idx):
        os.makedirs(f"sample_idx/{os.path.basename(args.gpt_ckpt)}/{idx}", exist_ok=True)
        print(f"generating samples for class {idx}")
        for j in tqdm(range(args.num_samples)):
            class_labels = [idx]
            c_indices = torch.tensor(class_labels, device=device)
            qzshape = [len(class_labels), args.codebook_embed_dim, latent_size, latent_size]

            t1 = time.time()
            cfg_scale = random.uniform(1.0, args.cfg_scale)
            index_sample = generate(
                gpt_model, c_indices, latent_size ** 2,
                cfg_scale=cfg_scale, cfg_interval=args.cfg_interval,
                temperature=args.temperature, top_k=args.top_k,
                top_p=args.top_p, sample_logits=True, 
                )
            with open(f"sample_idx/{os.path.basename(args.gpt_ckpt)}/{idx}/{j}_{cfg_scale}.npy", "wb") as f:
                np.save(f, index_sample.cpu().numpy())           
            


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--gpt-ckpt", type=str, default=None)
    parser.add_argument("--gpt-type", type=str, choices=['c2i', 't2i'], default="c2i", help="class-conditional or text-conditional")
    parser.add_argument("--from-fsdp", action='store_true')
    parser.add_argument("--cls-token-num", type=int, default=1, help="max token number of condition input")
    parser.add_argument("--precision", type=str, default='bf16', choices=["none", "fp16", "bf16"]) 
    parser.add_argument("--compile", action='store_true', default=False)
    parser.add_argument("--vq-model", type=str, choices=list(VQ_models.keys()), default="VQ-16")
    parser.add_argument("--vq-ckpt", type=str, default=None, help="ckpt path for vq model")
    parser.add_argument("--codebook-size", type=int, default=16384, help="codebook size for vector quantization")
    parser.add_argument("--codebook-embed-dim", type=int, default=8, help="codebook dimension for vector quantization")
    parser.add_argument("--image-size", type=int, choices=[256, 384, 512], default=384)
    parser.add_argument("--downsample-size", type=int, choices=[8, 16], default=16)
    parser.add_argument("--num-classes", type=int, default=1000)
    parser.add_argument("--cfg-scale", type=float, default=10.0)
    parser.add_argument("--cfg-interval", type=float, default=-1)
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--top-k", type=int, default=2000,help="top-k value to sample with")
    parser.add_argument("--temperature", type=float, default=1.0, help="temperature value to sample with")
    parser.add_argument("--top-p", type=float, default=1.0, help="top-p value to sample with")
    parser.add_argument("--start-idx", type=int, default=0)
    parser.add_argument("--end-idx", type=int, default=1000)
    parser.add_argument("--num-samples", type=int, default=30)
    args = parser.parse_args()
    main(args)